(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uarray.mli,v 1.22 2014/01/10 15:28:04 fclement Exp $ *)

(** {3 More array operations} *)

val length : 'a array -> Unatural.t;;
(** Return the length (number of elements) of the given array. *)

val make : Unatural.t -> 'a -> 'a array;;
(** [make n x] returns a fresh array of length [n],
   initialized with [x].
   All the elements of this new array are initially
   physically equal to [x] (in the sense of the [==] predicate).
   Consequently, if [x] is mutable, it is shared among all elements
   of the array, and modifying [x] through one of the array entries
   will modify all other entries at the same time.

   Raise [Invalid_argument] if [n > Sys.max_array_length].
   If the value of [x] is a floating-point number, then the maximum
   size is only [Sys.max_array_length / 2].*)

val init : Unatural.t -> (int -> 'a) -> 'a array
(** [init n f] returns a fresh array of length [n],
   with element number [i] initialized to the result of [f i].
   In other terms, [init n f] tabulates the results of [f]
   applied to the integers [0] to [n-1].

   Raise [Invalid_argument] if [n > Sys.max_array_length].
   If the return type of [f] is [float], then the maximum
   size is only [Sys.max_array_length / 2].*)

val is_empty : 'a array -> bool;;

val check_nonempty : Upervasives.function_name -> 'a array -> unit;;

(** {6 Filling arrays} *)

val set_sub : ('a -> 'a) -> 'a array -> int -> Unatural.t -> unit;;
(** [set_sub f a ofs len] applies function [f] in turn to the
   elements number [ofs] to [ofs + len - 1] of array [a] and stores the result
   to the corresponding location. It is equivalent to
   [a.(ofs) <- f a.(ofs);
    a.(ofs + 1) <- f a.(ofs + 1);
    ...;
    a.(ofs + len - 1) <- f a.(ofs + len - 1)].

   Raise [Invalid_argument] if [ofs] and [len] do not designate a valid
   subarray of [a]. *)

val set : ('a -> 'a) -> 'a array -> unit;;
(** [set f a] is equivalent to
   {!set_sub} [f a 0 (Array.length a)]. *)

val seti_sub :
    (int -> 'a -> 'a) -> 'a array -> int -> Unatural.t -> unit
;;
(** Same as {!set_sub}, but the function is applied to the index
   of the element as first argument, and the element itself as second argument.
   *)

val seti : (int -> 'a -> 'a) -> 'a array -> unit;;
(** [seti f a] is equivalent to
   {!seti_sub} [f a 0 (Array.length a)]. *)

val list_of_list_array : 'a list array -> 'a array list;;
(** *)

val list_to_list_array : 'a array list -> 'a list array;;
(** *)

val of_list1 : 'a Ulist1.t -> 'a array;;

(** {6 Iterators} *)

val iter_rails : ('a -> unit) array -> 'a array -> unit;;
(** [iter_rails fa a] applies components of function array [fa] in turn to
 all the elements of [a].
 It is equivalent to
 [f.(0) a.(0); f.(1) a.(1); ...;
  f.(Array.length fa - 1) a.(Array.length a - 1); ()].
  
  Raise [Invalid_arg] if the two arrays have different lengths. *)

val sub_iter : ('a -> unit) -> 'a array -> int -> Unatural.t -> unit;;
(** [sub_iter] is equivalent to [Array.iter] for the specified subarray.
  
  Raise [Invalid_argument] if the subarray specification is wrong. *)

val sub_iter_i : (int -> 'a -> unit) -> 'a array -> int -> Unatural.t -> unit;;
(** Similar to {!sub_iter}, but the function is applied to the index of
   the element as first argument, and the element itself as second argument. *)

val sub_fold_left :
  ('accu -> 'a -> 'accu) -> 'accu -> 'a array -> int -> Unatural.t -> 'accu
;;
(** [sub_fold_left] is equivalent to [Array.fold_left] for the specified
  subarray.
  
  Raise [Invalid_argument] if the subarray specification is wrong. *)

val fold_left_i :
  ('accu -> int -> 'a -> 'accu) -> 'accu -> 'a array -> 'accu
;;
(** Similar to [Array.fold_left], but the function is applied to the index of
 the element as second argument. *)

val fold_right_i :
  (int -> 'b -> 'accu -> 'accu) -> 'b array -> 'accu -> 'accu
;;
(** Similar to [Array.fold_right], but the function is applied to the index of
 the element as first argument. *)

val sub_fold_left_i :
  ('accu -> int -> 'a -> 'accu) ->
  'accu -> 'a array -> int -> Unatural.t -> 'accu
;;
(** Similar to {!sub_fold_left}, but the function is applied to the index of
 the element as second argument. *)

(** {6 Iterators on two arrays} *)

val iter2 : ('a -> 'b -> unit) -> 'a array -> 'b array -> unit;;
(** [iter2 f a b] applies function [f] in turn to all the elements
   of [a] and [b]. It is equivalent to
   [f a.(0) b.(0); f a.(1) b.(1); ...; f a.(n - 1) b.(n - 1); ()],
   where [n] is the length of arrays [a] and [b].
   
   Raise [Invalid_argument] if the two arrays have different lengths. *)

val map2 : ('a -> 'b -> 'c) -> 'a array -> 'b array -> 'c array;;
(** [map2 f a b] applies function [f] to all the elements of [a]
   and [b], and builds an array with the results returned by [f]:
   [[| f a.(0) b.(0); f a.(1) b.(1); ...; f a.(n - 1) b.(n - 1) |]],
   where [n] is the length of arrays [a] and [b].
   
   Raise [Invalid_argument] if the two arrays have different lengths. *)

val iter_i2 : (int -> 'a -> 'b -> unit) -> 'a array -> 'b array -> unit;;
(** Same as {!iter2}, but the function is applied to the index of the
   elements as first argument, and the elements themselves as second and third
   arguments. *)

val map_i2 : (int -> 'a -> 'b -> 'c) -> 'a array -> 'b array -> 'c array;;
(** Same as {!map2}, but the function is applied to the index of the
   elements as first argument, and the elements themselves as second and third
   arguments. *)

val fold_left2 :
  ('accu -> 'a -> 'b -> 'accu) -> 'accu -> 'a array -> 'b array -> 'accu
;;
(** [fold_left2 f x a b] computes
   [f (... (f (f x a.(0) b.(0)) a.(1) b.(1)) ...) a.(n - 1) b.(n - 1)],
   where [n] is the length of the arrays [a] and [b].
   
   Raise [Invalid_argument] if the two arrays have different lengths. *)

val fold_right2 :
  ('a -> 'b -> 'accu -> 'accu) -> 'a array -> 'b array -> 'accu -> 'accu
;;
(** [fold_right2 f a b x] computes
   [f a.(0) b.(0) (f a.(1) b.(1) ( ... (f a.(n - 1) b.(n - 1) x) ...))],
   where [n] is the length of the arrays [a] and [b].
   
   Raise [Invalid_argument] if the two arrays have different lengths. *)

val sub_iter2 :
  ('a -> 'b -> unit) -> 
  'a array -> int -> 'b array -> int -> Unatural.t -> unit
;;
(** [sub_iter2] is equivalent to {!iter2} for the specified subarray.
  
  Raise [Invalid_argument] if the subarray specification is wrong. *)

val sub_iter_i2 :
  (int -> 'a -> int -> 'b -> unit) ->
  'a array -> int -> 'b array -> int -> Unatural.t -> unit
;;
(** Similar to {!sub_iter2}, but the function is applied to the index of
   the elements as first and third arguments. *)

(** {6 Iterators on more arrays} *)

val iter3 :
  ('a -> 'b -> 'c -> unit) -> 'a array -> 'b array -> 'c array -> unit
;;
val iter4 :
  ('a -> 'b -> 'c -> 'd -> unit) ->
  'a array -> 'b array -> 'c array -> 'd array -> unit
;;
val map3 :
  ('a -> 'b -> 'c -> 'd) -> 'a array -> 'b array -> 'c array -> 'd array
;;
val map4 :
  ('a -> 'b -> 'c -> 'd -> 'e) ->
  'a array -> 'b array -> 'c array -> 'd array -> 'e array
;;
(** See {!iter2} and {!map2}.
   
   Raise [Invalid_argument] if argument arrays have different lengths. *)

(** {6 Iterators on list of arrays} *)

val maplist : ('a list -> 'a) -> 'a array list -> 'a array;;
(** [maplist f al] applies function [f] to all the elements of the
   items of the array list [al], and builds an array with the results returned
   by [f]:
   [[| f la.(0); f la.(1); ...; f la.(n - 1) |]],
   where [la] is the array of lists [[| [a1.(0); a2.(0); ...; ap.(0)];
   [a1.(1); ...; ap.(1)]; ...;[a1.(n - 1); ...; ap.(n - 1)] |]]
   and [n] is the length of arrays of the list [al].
   
   Raise [Invalid_argument] if the arrays of the list have different lengths. *)

(** {6 Array scanning} *)

val for_all : ('a -> bool) -> 'a array -> bool;;
(** [for_all p [|a1; ...; an|]] checks if all elements of the array
    satisfy the predicate [p]. That is, it returns
    [(p a1) && (p a2) && ... && (p an)]. *)

val exists : ('a -> bool) -> 'a array -> bool;;
(** [exists p [|a1; ...; an|]] checks if at least one element of
   the array satisfies the predicate [p]. That is, it returns
   [(p a1) || (p a2) || ... || (p an)]. *)

val find : ('a -> bool) -> 'a array -> 'a;;
(** [find p [|a1; ...; an|]] returns the first element of the array
   that satisfies the predicat [p] if any, otherwise raise [Not_found]. *)

val find_all : ('a -> bool) -> 'a array -> 'a list;;
(** [find_all p [|a1; ...; an|]] returns all the elements of the array
   that satisfies the predicat [p] if any (in reverse order), otherwise
   returns [[]]. *)

val for_all2 : ('a -> 'b -> bool) -> 'a array -> 'b array -> bool;;

val sub_find : ('a -> bool) -> 'a array -> int -> Unatural.t -> 'a;;
(** [sub_find] is equivalent to {!find} for the specified subarray.
  
  Raise [Invalid_argument] if the subarray specification is wrong. *)

val sub_find_all : ('a -> bool) -> 'a array -> int -> Unatural.t -> 'a list;;
(** [sub_find_all] is equivalent to {!find_all} for the specified subarray.
  
  Raise [Invalid_argument] if the subarray specification is wrong. *)

(** {6 Sorted arrays} *)

val is_sorted : ('a -> 'a -> bool) -> 'a array -> bool;;

val is_strictly_increasing : 'a array -> bool;;
val check_strictly_increasing : Upervasives.function_name -> 'a array -> unit;;

val sorted_array_find : 'a array -> 'a -> Unatural.t;;
