(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ugraph.ml,v 1.2 2014/01/10 15:28:04 fclement Exp $ *)

let g_invalid_arg = Upervasives.invalid_arg "Ugraph";;

type 'a index = 'a Unonempty_set.index
and 'a uindex = 'a Unonempty_set.uindex
;;

type 'a domain = 'a Unonempty_set.t;;

type 'a sub_domain = 'a Unonempty_set.nonempty_subset;;

type 'a images =
  | Constant of 'a
  | Images of 'a array
;;

type ('a, 'b) t =
  {
   domain : 'a domain;
   images : 'b images;
  }
;;

let same g1 g2 = g1 == g2;;

(* Construction functions. *)

let constant dom y = { domain = dom; images = Constant y; };;

let from_array dom ims =
  if Unonempty_set.cardinal dom <> Uarray.length ims then
    g_invalid_arg "from_array" "arguments have different sizes" else
  { domain = dom; images = Images ims; }
;;

let from_fun dom f =
  let uindices = Unonempty_set.uindices dom in
  let y0 = f Uindex._0_ uindices.(0) in
  let constant = ref true in
  let ff i uindex =
    if i = 0 then y0 else
    let y = f (Uindex.of_int i) uindex in
    if not !constant then y else
    if y = y0 then y0 else begin constant := false; y end in
  let ims = Array.mapi ff uindices in
  let images = if !constant then Constant y0 else Images ims in
  { domain = dom; images = images; }
;;

let of_matrix m =
  let (height, width) = Umatrix.size m in
  let dom =
    Unonempty_set.of_rectangle
      (Unatural.to_int height) (Unatural.to_int width) in
  let ims = Umatrix.to_array m in
  { domain = dom; images = Images ims; }
;;

(* Functions on graphs. *)

let image g =
  match g.images with
  | Constant y -> (fun _index -> y)
  | Images ims -> (fun index -> ims.(Uindex.to_int index))
;;

let restriction
    ({ domain = dom; images = images; }  as g)
    { Unonempty_set.support = sup;
      indices = indices;
      nonempty_set = sub_dom; } =
  if sup != dom then
    g_invalid_arg "restriction" "sub-domain is not a subset of domain" else
  match images with
  | Constant y -> constant sub_dom y
  | Images _ -> from_array sub_dom (Array.map (image g) indices)
;;

(* Iterators. *)

let map_images f g =
  match g.images with
  | Constant y -> { g with images = Constant (f y); }
  | Images ims -> { g with images = Images (Array.map f ims); }
;;
