(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulistn.ml,v 1.8 2014/01/17 19:19:30 fclement Exp $ *)

let ln_invalid_arg fun_name = Upervasives.invalid_arg "Ulistn" fun_name "";;

open Ulistn_spec;;

module Succ_structure =
  functor (Listpn_structure : LISTN_STRUCTURE) ->
struct
  type 'a listpn = 'a Listpn_structure.t;;
  type 'a listn = 'a * 'a listpn;;
  type 'a t = 'a listn;;
  
  let to_listpn (h, lpn) = Listpn_structure.cons h lpn;;
  let of_listpn lpn =
    (Listpn_structure.hd lpn,
     Listpn_structure.of_listpn (Listpn_structure.tl lpn))
  ;;
  
  let min_length = Unatural.succ Listpn_structure.min_length;;
  
  let to_list (h, lpn) = h :: Listpn_structure.to_list lpn;;
  let of_list = function
    | [] -> ln_invalid_arg "of_list"
    | h :: lpn -> (h, Listpn_structure.of_list lpn)
  ;;
  
  let cons x (h, lpn) = (x, Listpn_structure.cons h lpn);;
  let hd (h, _) = h;;
  let tl (_, l) = l;;
end
;;

module Make =
  functor (Listn_structure : LISTN_STRUCTURE) ->
struct
  include Listn_structure;;
  
  let lift_dom f ln = f (to_list ln);;
  let lift_codom f x = of_list (f x);;
  let lift f = lift_codom (lift_dom f);;
  let lift_dom2 f ln1 ln2 = f (to_list ln1) (to_list ln2);;
  let lift_codom2 f x1 x2 = of_list (f x1 x2);;
  let lift2 f = lift_codom2 (lift_dom2 f);;
  
  let proj_codom f x = to_list (f x);;
  let proj_codom_i f i = proj_codom (f i);;
  let proj_codom2 f x1 x2 = to_list (f x1 x2);;
  let proj_codom2_i f i = proj_codom2 (f i);;
  
  let length ln = lift_dom Ulist.length ln;;
  let first n ln = lift_dom (Ulist.first n) ln;;
  let last n ln = lift_dom (Ulist.last n) ln;;
  let nth ln n = lift_dom (fun l -> Ulist.nth l n) ln;;
  let rank e = lift_dom (Ulist.rank e);;
  let rev ln = lift Ulist.rev ln;;
  let append ln1 ln2 = lift2 Ulist.append ln1 ln2;;
  let rev_append ln1 ln2 = lift2 Ulist.rev_append ln1 ln2;;
  let init n f = of_list (Ulist.init n f);;
  let rev_init n f = of_list (Ulist.rev_init n f);;
  let replace i x = lift (Ulist.replace i x);;
  let rev_replace i x = lift (Ulist.rev_replace i x);;
  let replace_by_list i = lift2 (Ulist.replace_by_list i);;
  let rev_replace_by_list i = lift2 (Ulist.rev_replace_by_list i);;
  let insert i = lift2 (Ulist.insert i);;
  let rev_insert i = lift2 (Ulist.rev_insert i);;
  
  (* Iterators. *)
  let iter f = lift_dom (Ulist.iter f);;
  let iter_i_shift f i = lift_dom (Ulist.iter_i_shift f i);;
  let iter_i f = iter_i_shift f 0;;
  let map f = lift (Ulist.map f);;
  let map_i_shift f i = lift (Ulist.map_i_shift f i);;
  let map_i f = map_i_shift f 0;;
  let rev_map f = lift (Ulist.rev_map f);;
  let rev_map_i_end_shift accu f i = lift (Ulist.rev_map_i_end_shift accu f i);;
  let rev_map_i f = rev_map_i_end_shift [] f 0;;
  let flat_map_end accu f = lift (Ulist.flat_map_end accu (proj_codom f));;
  let flat_map f = flat_map_end [] f;;
  let concat ll = flat_map Upervasives.id ll;;
  let flatten = concat;;
  let flat_map_i_end_shift accu f i =
    lift (Ulist.flat_map_i_end_shift accu (proj_codom_i f) i)
  ;;
  let flat_map_i f = flat_map_i_end_shift [] f 0;;
  let rev_flat_map_end accu f =
    lift (Ulist.rev_flat_map_end accu (proj_codom f))
  ;;
  let rev_flat_map f = rev_flat_map_end [] f;;
  let rev_flat_map_i_end_shift accu f i =
    lift (Ulist.rev_flat_map_i_end_shift accu (proj_codom_i f) i)
  ;;
  let rev_flat_map_i f = rev_flat_map_i_end_shift [] f 0;;
  let fold_left f accu = lift_dom (Ulist.fold_left f accu);;
  let fold_left_i_shift f i accu =
    lift_dom (Ulist.fold_left_i_shift f i accu)
  ;;
  let fold_left_i f = fold_left_i_shift f 0;;
  let fold_right f ln accu =
    lift_dom (fun l -> Ulist.fold_right f l accu) ln
  ;;
  let fold_right_i_shift f i ln accu =
    lift_dom (fun l -> Ulist.fold_right_i_shift f i l accu) ln
  ;;
  let fold_right_i f = fold_right_i_shift f 0;;
  let rev_fold_left_env f (env, accu) ln =
    let f (env, accu) x =
      let (env, accu) = f (env, accu) x in
      (env, to_list accu) in
    let (env, accu) = lift_dom (Ulist.fold_left f (env, accu)) ln in
    (env, of_list accu)
  ;;
  let fold_left_env f (env, accu) ln =
    let (env, accu) = rev_fold_left_env f (env, Ulist.rev accu) ln in
    (env, rev accu)
  ;;
  let fold_left_env_init f env = fold_left_env f (env, []);;
  
  (* Iterators on two lists with minimum length. *)
  let iter2 f = lift_dom2 (Ulist.iter2 f);;
  let iter2_i_shift f i = lift_dom2 (Ulist.iter2_i_shift f i);;
  let iter2_i f = iter2_i_shift f 0;;
  let map2 f = lift2 (Ulist.map2 f);;
  let map2_i_shift f i = lift2 (Ulist.map2_i_shift f i);;
  let map2_i f = map2_i_shift f 0;;
  let rev_map2 f = lift2 (Ulist.rev_map2 f);;
  let rev_map2_i_end_shift accu f i =
    lift2 (Ulist.rev_map2_i_end_shift accu f i)
  ;;
  let rev_map2_i f = rev_map2_i_end_shift [] f 0;;
  let flat_map2_end accu f = lift2 (Ulist.flat_map2_end accu (proj_codom2 f));;
  let flat_map2 f = flat_map2_end [] f;;
  let flat_map2_i_end_shift accu f i =
    lift2 (Ulist.flat_map2_i_end_shift accu (proj_codom2_i f) i)
  ;;
  let flat_map2_i f = flat_map2_i_end_shift [] f 0;;
  let rev_flat_map2_end accu f =
    lift2 (Ulist.rev_flat_map2_end accu (proj_codom2 f))
  ;;
  let rev_flat_map2 f = rev_flat_map2_end [] f;;
  let rev_flat_map2_i_end_shift accu f i =
    lift2 (Ulist.rev_flat_map2_i_end_shift accu (proj_codom2_i f) i)
  ;;
  let rev_flat_map2_i f = rev_flat_map2_i_end_shift [] f 0;;
  let fold_left2 f accu = lift_dom2 (Ulist.fold_left2 f accu);;
  let fold_left2_i_shift f i accu =
    lift_dom2 (Ulist.fold_left2_i_shift f i accu)
  ;;
  let fold_left2_i f = fold_left2_i_shift f 0;;
  let fold_right2 f ln1 ln2 accu =
    lift_dom2 (fun l1 l2 -> Ulist.fold_right2 f l1 l2 accu) ln1 ln2
  ;;
  let fold_right2_i_shift f i ln1 ln2 accu =
    lift_dom2 (fun l1 l2 -> Ulist.fold_right2_i_shift f i l1 l2 accu) ln1 ln2
  ;;
  let fold_right2_i f = fold_right2_i_shift f 0;;
  let rev_fold_left2_env f (env, accu) ln1 ln2 =
    let f (env, accu) x1 x2 =
      let (env, accu) = f (env, accu) x1 x2 in
      (env, to_list accu) in
    let (env, accu) = lift_dom2 (Ulist.fold_left2 f (env, accu)) ln1 ln2 in
    (env, of_list accu)
  ;;
  let fold_left2_env f (env, accu) lpn1 lpn2 =
    let (env, accu) = rev_fold_left2_env f (env, Ulist.rev accu) lpn1 lpn2 in
    (env, rev accu)
  ;;
  let fold_left2_env_init f env = fold_left2_env f (env, []);;
  
  (* Scanning. *)
  let for_all p = lift_dom (Ulist.for_all p);;
  let exists p = lift_dom (Ulist.exists p);;
  let for_all2 p = lift_dom2 (Ulist.for_all2 p);;
  let exists2 p = lift_dom2 (Ulist.exists2 p);;
  let mem a = exists (fun x -> x = a);;
  let memq a = exists (fun x -> x == a);;
  
  (* Searching. *)
  let find p = lift_dom (Ulist.find p);;
  let find_all p = lift_dom (Ulist.find_all p);;
  let filter = find_all;;
  let partition p = lift_dom (Ulist.partition p);;
  
  (* Association lists with minimum length. *)
  let assoc k = lift_dom (Ulist.assoc k);;
  let assq k = lift_dom (Ulist.assq k);;
  let mem_assoc k = exists (fun (a, _) -> k = a);;
  let mem_assq k = exists (fun (a, _) -> k == a);;
  let remove_assoc k = lift_dom (Ulist.remove_assoc k);;
  let remove_assq k = lift_dom (Ulist.remove_assq k);;
  let compare_assoc cmp = lift_dom (Ulist.compare_assoc cmp);;
  let minus_pervasives_compare x y = - Pervasives.compare x y;;
  let min_assoc ln = compare_assoc Pervasives.compare ln
  and max_assoc ln = compare_assoc minus_pervasives_compare ln
  ;;
  let compare_fst_assoc = compare_assoc
  and min_fst_assoc = min_assoc
  and max_fst_assoc = max_assoc
  ;;
  let compare_snd_assoc cmp = lift_dom (Ulist.compare_snd_assoc cmp);;
  let min_snd_assoc ln = compare_snd_assoc Pervasives.compare ln
  and max_snd_assoc ln = compare_snd_assoc minus_pervasives_compare ln
  ;;
  
  (* Lists of pairs with minimum length. *)
  let lift_pair f ln =
    let (res1, res2) = lift_dom f ln in
    (of_list res1, of_list res2)
  ;;
  let split ln = lift_pair Ulist.split ln;;
  let combine ln1 ln2 = lift2 Ulist.combine ln1 ln2;;
  
  (* Sorting. *)
  let sort cmp = lift (Ulist.sort cmp);;
  let stable_sort cmp = lift (Ulist.stable_sort cmp);;
  let fast_sort cmp = lift (Ulist.fast_sort cmp);;
  let merge cmp = lift2 (Ulist.merge cmp);;
  let of_range pred l r = of_list (Ulist.of_range pred l r);;
end
;;

module Succ =
  functor (Listn : LISTN) ->
  Make (Succ_structure (Listn))
;;
