(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: unonempty_set.ml,v 1.3 2014/01/10 15:28:04 fclement Exp $ *)

let nes_invalid_arg = Upervasives.invalid_arg "Unonempty_set";;
let u_check_index fun_name = Uuniverse.check_index "Unonempty_set" fun_name;;

type 'a uindex = 'a Uuniverse.index;;

type 'a index = 'a uindex uindex;;

type 'a t =
  {
   universe : 'a Uuniverse.t;
   uindices : 'a uindex Uuniverse.t;
  }
;;

(* Construction functions. *)

let of_array u uindices =
  let len = Uarray.length uindices in
  if len = Unatural._0_ then nes_invalid_arg "of_array" "empty array" else
  begin
    Array.sort Pervasives.compare uindices;
    let last_uindex = uindices.(pred (Unatural.to_int len)) in
    u_check_index "of_array" u last_uindex;
    { universe = u; uindices = Uuniverse.of_array uindices; }
  end
;;

let singleton u uindex = of_array u [| uindex; |];;

let full u = { universe = u; uindices = Uuniverse.indices u; };;

let rec ulist1_last = function
  | (x, []) -> x
  | (_, x :: l) -> ulist1_last (x, l)
;;

let unsafe_of_list u uindices =
  { universe = u; uindices = Uuniverse.of_list uindices; }
;;

let of_list u uindices =
  let uindices = Ulist1.sort Pervasives.compare uindices in
  let last_uindex = ulist1_last uindices in
  u_check_index "of_list" u last_uindex;
  unsafe_of_list u uindices
;;

let of_rectangle h w = full (Uuniverse.of_rectangle h w);;

let of_natural_rectangle h w = full (Uuniverse.of_natural_rectangle h w);;

(* Indices and elements. *)

let cardinal nes = Uuniverse.cardinal nes.uindices;;

let valid_index nes index = Uindex.to_natural index < cardinal nes;;

let check_index fun_name nes index =
  if not (valid_index nes index) then
    nes_invalid_arg fun_name
      (Printf.sprintf "index %i out of set of %d element(s)"
         (Uindex.to_int index)
         (Unatural.to_int (cardinal nes)))
;;

let mem nes uindex = Uuniverse.mem nes.uindices uindex;;

let choice nes =
  Uuniverse.element nes.uindices (Uindex.of_natural Unatural._0_)
;;

let uindex nes index =
  check_index "indice" nes index;
  Uuniverse.element nes.uindices index
;;

let uindices nes = Uuniverse.elements nes.uindices;;

let element nes index = Uuniverse.element nes.universe (uindex nes index);;

let elements nes = Uuniverse.sub_elements nes.universe (uindices nes);;

let add nes uindex =
  u_check_index "add_to_nonempty" nes.universe uindex;
  if not (mem nes uindex) then Uuniverse.add nes.uindices uindex
;;

(* Iterators. *)

let same_universe nes1 nes2 = nes1.universe == nes2.universe;;

let check_same_universe fun_name nes1 nes2 =
  if not (same_universe nes1 nes2) then
    nes_invalid_arg fun_name "universe clash"
;;

let fold_left f accu nes = Uuniverse.fold_left f accu nes.uindices;;

let fold_left_i f accu nes = Uuniverse.fold_left_i f accu nes.uindices;;

let union nes1 nes2 =
  check_same_universe "union" nes1 nes2;
  let (smaller, greater) =
    if cardinal nes1 < cardinal nes2 then (nes1, nes2) else (nes2, nes1) in
  let uindices =
    fold_left
      (fun accu i -> Uuniverse.add accu i; accu)
      (Uuniverse.copy greater.uindices) smaller in
  { greater with uindices = uindices; }
;;

let iter f nes = Uuniverse.iter f nes.uindices;;

let iter_i f nes = Uuniverse.iter_i f nes.uindices;;

let map_indices f nes = Uuniverse.map_elements f nes.uindices;;

let map f nes u = of_array u (map_indices f nes);;

let find p nes = Uuniverse.find p nes.uindices;;

let find_all p nes = Uuniverse.find_all p nes.uindices;;

let filter = find_all;;

(* Nonempty subsets. *)

type 'a nonempty_subset =
  {
   support : 'a t;
   indices : 'a index array;
   nonempty_set : 'a t;
  }
;;

let is_full nes = cardinal nes = Uuniverse.cardinal nes.universe;;

let check_full fun_name nes =
  if not (is_full nes) then nes_invalid_arg fun_name "nonempty set is not full"
;;

let same_full nes1 nes2 =
  same_universe nes1 nes2 && is_full nes1 && is_full nes2
;;

let check_same_full fun_name nes1 nes2 =
  check_same_universe fun_name nes1 nes2;
  check_full fun_name nes1;
  check_full fun_name nes2
;;

let (full_index_of_uindex : 'a uindex -> 'a index) =
  fun uindex -> Uindex.of_natural (Uindex.to_natural uindex)
;;

(* We use here the fact that the full set, containing all the elements of the
  universe, is numbered in the same order as the universe.
  Hence, the use of magic function index_of_uindex.
  If efficiency is an issue, consider using Obj.magic on the whole array of
  indices. *)
let subset_of_full full nes =
  check_full "subset_of_full" full;
  check_same_universe "subset_of_full" full nes;
  let indices =
    Array.map full_index_of_uindex (Uuniverse.elements nes.uindices) in
  { support = full; indices = indices; nonempty_set = nes; }
;;

type 'a dichotomy = 'a nonempty_subset * 'a nonempty_subset;;

let dichotomy p nes =
  let dicho (tl, fl) index uindex =
    if p index uindex then (index :: tl, fl) else (tl, index :: fl) in
  let (tl, fl) =
    match fold_left_i dicho ([], []) nes with
    | ([], _) | (_, []) -> nes_invalid_arg "dichotomy" "empty subset"
    | (tx :: tl, fx :: fl) -> ((tx, tl), (fx, fl)) in
  let uindex index = Uuniverse.element nes.uindices index in
  let (utl, ufl) = (Ulist1.rev_map uindex tl, Ulist1.rev_map uindex fl) in
  let (tnes, fnes) =
    (unsafe_of_list nes.universe utl, unsafe_of_list nes.universe ufl) in
  ({ support = nes; indices = Uarray.of_list1 tl; nonempty_set = tnes; },
   { support = nes; indices = Uarray.of_list1 fl; nonempty_set = fnes; })
;;
